#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include "mytypes.h"
#include "dis.h"
#include "symdef.h"

#define MAXADDR 65536l
#define VERSION "dis96 v. 1.05 (" __DATE__ ") (c) 1998-2003 P. Heitzer"
#define INST 'I'
#define DUMP 'B'

typedef char SFR[16];

extern Word maxaddr;
extern Byte *mem;
extern SYMBOL symlist[];
extern SFR sfrname[256];
extern int symcount;

void changeSFRName(Byte num, char *newname);
int doList(Word addr);
void doListControl(Word addr,Byte mode);
int doDisp(Word addr);
int getcmd(Word *addr);
int doHelp(void);
int saveControlTable(char *f);
int loadControlTable(char *f);
int doXDisp(Word addr);
int isInstr(Word addr);

static Byte *listControl,listMode=INST;
static char cmd;

int doHelp(void)
{
printf(VERSION "\nAvailable commands are: \n");
printf(" h: this help\n");
printf(" [n] d [addr] : dump next 16*n bytes as DB ..\n");
printf(" [n] l [addr] : disassemble next n instructions\n");
printf(" [n] x [addr] : dump next 16*n bytes as hex digits\n");
printf(" c addr {i|b} : set list control for address addr to i(nstruction) or b(yte)\n");
printf(" = addr name : assign addr symbolic name\n");
printf(" . byte name : assign SFR byte symbolic name\n");
printf(" r filename : read control table from file\n");
printf(" s filename : save control table to file\n");
printf(" q : quit\n");
printf(" at start n defaults to 10\n");
}

int isInstr(Word addr)
{
if (listControl==NULL)
  return 1;
if (listControl[addr]==DUMP)
  { listMode=DUMP; return 0; }
if (listControl[addr]==INST)
  { listMode=INST; return 1;}
else
  return listMode==INST;
}

int doList(Word addr)
{
Byte *buf=mem+addr;
int len;
if (isInstr(addr))
  len=instr(addr);
else
  len=doXDisp(addr);
return len;
}

int doDisp(Word addr)
{
Byte *buf=mem+addr;
char tmp[3];

int len=16,i;
printf("%s DB ",symstr(addr));
for (i=0;i<len;i++)
  {
  if (cmd=='L')
    if (isInstr(addr+i))
      {
      printf("\n");
      return i;
      }
  if (isprint(buf[i]))
    printf("'%c'",buf[i]);
  else
    {
    sprintf(tmp,"%x",buf[i]);
    if (*tmp>'9')
      printf("0");
    printf("%s%s",tmp,buf[i]>9 ? "h":"");
    }
  if (i<len-1)
    printf(",");
  }
printf("\n");
return len;
}

int doXDisp(Word addr)
{
Byte *buf=mem+addr;
int len=16,i;

printsym(addr);

for (i=0;i<16;i++)
  {
  /* if (isInstr(addr+i) && len==16)
    len=i; */
  if (i<len)
    printf("%02X ",buf[i]);
  else
    printf("   ");
  }
for (i=0;i<len;i++)
  printf("%c",isprint(buf[i])? buf[i]:'.');
printf("\n");
return len;
}

void doListControl(Word addr,Byte mode)
{
if (listControl==NULL)
  if ((listControl=(Byte *)calloc(MAXADDR,1))==NULL)
    return;
switch (toupper(mode)) {
  case INST: listControl[addr]=INST; break;
  case DUMP: listControl[addr]=DUMP; break;
  }
}

int saveControlTable(char *fn)
{
long i;
FILE *f;
char tmp[4];

if (f=fopen(fn,"w+")) {
  if (listControl!=NULL)
    for (i=0;i<MAXADDR;i++)
      if (listControl[i]==DUMP)
	fprintf(f,"C %04lx B\n",i);
      else
	if (listControl[i]==INST)
	  fprintf(f,"C %04lx I\n",i);
  for (i=0;i<symcount;i++)
    fprintf(f,"= %04lx %s\n",symlist[i].addr,symlist[i].name);
  for (i=0;i<256;i++) {
	  sprintf(tmp,"%02X",i);
		if (strcmp(tmp,sfrname[i]))
      fprintf(f,". %02x %s\n",i,sfrname[i]);
		}	
  fclose(f);
  return 0;
  }
return 1;
}

int loadControlTable(char *fn)
{
FILE *f;
char cmdline[80],*p1,*p2,*p3;
Word addr;

if ((f=fopen(fn,"r"))==NULL)
  return -1;
for (;;) {
  if (fgets(cmdline,80,f)==NULL)
    break;
  if (p1=strtok(cmdline," \n"))
    if (p2=strtok(NULL," \n"))
      if (p3=strtok(NULL," \n"))
	{
	addr=strtoul(p2,NULL,16);
	switch (toupper(*p1)) {
	  case 'C' : doListControl(addr,*p3); break;
	  case '=' : addsym(p3,addr); break;
	  case '.' : changeSFRName(addr,p3); break;
	  }
	}
  }
fclose(f);
return 0;
}

int getcmd(Word *addr)
{
char cmdline[80],*cp,*p1=NULL,*p2=NULL,*p3=NULL;
Word newaddr;
SYMBOL *sym;
int i;
static int n=10;
static char lastcmd=' ';

*cmdline=0;

cp=fgets(cmdline,80,stdin);
if (p1=strtok(cp," \n"))
  {
  if (isdigit(*p1)) {
    if (!(n=atoi(p1))) n=10;
    p1=strtok(NULL," \n");
    }
  p2=strtok(NULL," \n");
  }
if (p2)
  {
  if (sym=findsym(p2,0))
    newaddr=sym->addr;
  else
    newaddr=strtoul(p2,NULL,16);
  p3=strtok(NULL," \n");
  }
if (p1) {
  cmd=toupper(*p1);
  if (strchr("DLX",cmd))
    {
    if (p2)
      *addr=newaddr;
    lastcmd=cmd;
    }
  }

if (p1==NULL || *p1==0) cmd=lastcmd;
switch (cmd) {
  case 'C' : if (p3!=NULL) doListControl(newaddr,*p3); break;
  case '=' : if (p3!=NULL) addsym(p3,newaddr); break;
  case '.' : if (p3!=NULL) changeSFRName(newaddr & 0xff,p3); break;
  case 'D' : if (*addr>=maxaddr) break;
			 for (i=0;i<n;i++)
			   *addr+=doDisp(*addr);
			 break;
  case 'H' : doHelp(); break;
  case 'L' : if (*addr>=maxaddr) break;
	     for (i=0;i<n;i++)
	       *addr+=doList(*addr);
	     break;
  case 'Q' : if (listControl) free(listControl); return 0;
  case 'R' : if (p2!=NULL) loadControlTable(p2); break;
  case 'S' : if (p2!=NULL) saveControlTable(p2); break;
  case 'X' : if (*addr>=maxaddr) break;
	     for (i=0;i<n;i++)
	       *addr+=doXDisp(*addr);
	     break;
  }
return 1;
}
