#include <stdio.h>
#include "mytypes.h"
#include "opcodes.h"
#include "globals.h"

extern OPCODE opcodes[];
extern int nrofmnemonics;
char outs[64],outm[16],buf[64];

typedef struct {
	Byte regnum;
	char *regstr;
	} REGSTR;
	
typedef char SFR[16];

/* symbolic names for SFRs */

REGSTR regstr[]={
	0x2, "AD_PORT",
	0x4, "HSI_TIME",
	0x6, "HSI_STAT",
	0x7, "SBUF",
	0x8, "INT_MASK",
	0x9, "INT_PEND",
	0xa, "TIMER1",
	0xc, "TIMER2",
	0xe, "PORT0",
	0xf, "PORT1",
	0x10, "PORT2",
	0x11, "SP_STAT",
	0x18, "SP",
	0, NULL
	};

SFR sfrname[256];

void initSFRNames() {
int i;
REGSTR *rs; 
for (i=0,rs=regstr;rs->regstr;i++,rs++)
  strcpy(sfrname[rs->regnum],rs->regstr);
for (i=0;i<256;i++)
  if (sfrname[i][0]==0)
	  sprintf(sfrname[i],"%02X",i);
}

void changeSFRName(Byte num, char *newname) {
strcpy(sfrname[num],newname);
}
char *reg2str0(Byte num) {
static char rbuf[128];
char *buf;
static int bi=0;
int i;

for (i=0;i<sizeof(regstr)/sizeof(REGSTR);i++)
  if (regstr[i].regnum==num)
	  return regstr[i].regstr;
bi=(bi+1) % 4;
buf=rbuf+bi*32;
sprintf(buf,"%02X",num);
return buf;
}

char *reg2str(Byte num) {
return sfrname[num];
}

OPCODE *findmnemonic(Byte *PC)
{
short i;
unsigned h=0;
Byte b;
OPCODE *op;

if (*PC==0xfe)	/* prefix for signed mul/div */
  {
  h=0xfe00;
  PC++;
  }

b=*PC;
h|=b;
PC++;

for (i=0,op=opcodes;i<nrofmnemonics;i++,op++)
  if ((op->opcode)==h)	     /* test if implied addressing */
    return op;

for (i=0,op=opcodes;i<nrofmnemonics;i++,op++)
  if (op->opcode==(h & 0xfffc)) /* test if normal addressing */
    return op;

for (i=0,op=opcodes;i<nrofmnemonics;i++,op++)
  if ((op->opcode&0xffdf)==(h&0xffdc))	/* test if normal addressing with 3 operands */
    if ((op->amode==NOR3) || (op->amode==NOR83))
      return op;

for (i=0,op=opcodes;i<nrofmnemonics;i++,op++)
  if (op->opcode==(h&0xfff8))	/* test if abs or rel 11 bit or bit # addressing */
    if ((op->amode==REL11) || (op->amode==A11) ||(op->amode==BIT))
	return op;

return 0;
}

int doNormal(ATYPE atype,Word PC)
{
Byte amode,*p=mem+PC,is3=0,is8=0,longInd,ireg;

amode=*p & 3;
is3=(atype==NOR3 || atype==NOR83) && !(*p&0x20);
is8=(atype==NOR8 || atype==NOR83);

switch (amode) {
  case 0: // reg,reg  src dst , sr2 dst src1
       if (!is3)
	 sprintf(outs,"%s,%s",reg2str(*(p+2)),reg2str(*(p+1)));
       else
	 sprintf(outs,"%s,%s,%s",reg2str(*(p+2)),reg2str(*(p+3)),
	 reg2str(*(p+1)));
       return 3+is3;

  case 1: // immediate
       if (!is3)
	 {
	 if (is8)
	   sprintf(outs,"%s,#%02X",reg2str(*(p+2)),*(p+1));
	 else
	   sprintf(outs,"%s,#%02X%02X",reg2str(*(p+3)),*(p+2),*(p+1));
	 return 4-is8;
	 }
       else
	 {
	 if (is8)
	   sprintf(outs,"%s,%s,#%02X",reg2str(*(p+2)),reg2str(*(p+3)),*(p+1));
	 else
	   sprintf(outs,"%s,%s,#%02X%02X",reg2str(*(p+3)),reg2str(*(p+4)),*(p+2),*(p+2));
	 return 5-is8;
	 }

  case 2: // indirect
       sprintf(outs,"%s,[%s]%s",reg2str(*(p+2)),
			 reg2str(*(p+1)&0xfe),*(p+1)&1?"+":"");
       return 3;

  case 3: // indexed  opc ireg mem(2,32) dest(3,4) src2(4,5)
       ireg=*(p+1)&0xfe; longInd=*(p+1)&1;
       if (!is3)
	 {
	 if (longInd)
	   sprintf(outs,"%s,%s[%s]",reg2str(*(p+4)),symstr( (Word)*(p+3)<<8|*(p+2)), 
		 reg2str(ireg));
	 else
	   sprintf(outs,"%s,%02X[%s]",reg2str(*(p+3)),*(p+2),reg2str(ireg));
	 return 4+longInd;
	 }
       else
	 {
	 if (longInd)
	   sprintf(outs,"%s,%s,%02X%02X[%s]",reg2str(*(p+4)),reg2str(*(p+5)),
		 *(p+3),*(p+2),reg2str(ireg));
	 else
	   sprintf(outs,"%s,%s,%02X[%s]",reg2str(*(p+3)),
		 reg2str(*(p+5)),*(p+4),reg2str(ireg));
	 return 5+longInd;
	 }
  }
return 2;
}

int doPushPop(ATYPE atype,Word PC)
{
Byte amode,*p=mem+PC,longInd,ireg;

amode=*p & 3;

switch (amode) {
  case 0: // reg,reg
    sprintf(outs,"%s",reg2str(*(p+1)));
    return 2;

  case 1: // immediate
    sprintf(outs,"#%02X%02X",*(p+2),*(p+1));
    return 3;

  case 2: // indirect
    sprintf(outs,"[%s]%s",reg2str(*(p+1)&0xfe),*(p+1)&1?"+":"");
    return 2;

  case 3: // indexed
    ireg=*(p+1)&0xfe; longInd=*(p+1)&1;
    if (longInd)
      sprintf(outs,"%02X%02X[%s]",*(p+3),*(p+2),reg2str(ireg));
    else
      sprintf(outs,"%02X[%s]",*(p+2),reg2str(ireg));
    return 3+longInd;
    }

return 2;
}

int doShift(ATYPE atype,Word PC)
{
Byte cnt,*p=mem+PC;

cnt=*(p+1);

switch (atype) {
  case SHIFT : sprintf(outs,"%s,%s%02X",reg2str(*(p+2)),(cnt<16) ? "\0":"R",cnt); break;
  case SHIFT8 : sprintf(outs,"%s,%s%02X",reg2str(*(p+2)),(cnt<8) ? "\0":"R",cnt); break;
  case SHIFT32 : sprintf(outs,"%s,%s%02X",reg2str(*(p+2)),(cnt<32) ? "\0":"R",cnt); break;
  }
return 3;
}

int disassemble(Word PC)
{
OPCODE *op;
char tmp[64];
short n=1,disp,prefix=0,i;
Byte *p;
Word addr;

p=(mem+PC);
/* fprintf(outf,"%04X: ",PC,*p); */
if (op=findmnemonic(p))
  {
  sprintf(outm,"%s ",op->mnemonic);
  for (i=0;i<32;i++) outs[i]=0;
  if (*p==0xfe)
    {
    p++;
    prefix=1;
    PC++;
    }
  switch (op->amode) {
    case IMP: break;
    case NOR:
    case NOR8:
    case NOR3:
    case NOR83: n=doNormal(op->amode,PC); break;
    case PUSHPOP: n=doPushPop(op->amode,PC); break;
    case REL: n=2; disp=(signed char)*(p+1); sprintf(outs,"%s",symstr(PC+disp+2)); break;
    case REL11: n=2; disp=((*p & 7)<<8)|*(p+1); if (disp & 0x400) disp |=0xf800;
						sprintf(outs,"%s",symstr((Word) (PC+disp+2))); break;
    case REL16: n=3; disp=(*(p+1))|(*(p+2)<<8); sprintf(outs,"%s",symstr((Word) (PC+disp+3))); break;
//  case A11: n=2; addr=((*p & 7)<<8)|*(p+1); sprintf(outs,"%04X",addr); break;
    case REG:
    case REG8: n=2; sprintf(outs,"%s",reg2str(*(p+1))); break;
    case IREG: n=2; sprintf(outs,"[%s]",reg2str(*(p+1))); break;
    case DJNZ: n=3; disp=(signed char)*(p+2);
		    sprintf(outs,"%s,%s",reg2str(*(p+1)),symstr(PC+disp+3)); break;
    case BIT:  n=3; disp=(signed char)*(p+2);
	       sprintf(outs,"%s,%d %s",reg2str(*(p+1)),*p&7,symstr(PC+3+disp));
	       break;
    case SHIFT:
    case SHIFT8:
    case SHIFT32: n=doShift(op->amode,PC); break;
    }
	/*
  for (i=0;i<8;i++)
    if (i<n+prefix)
      fprintf(outf,"%02X",mem[PC-prefix+i]);
    else
      fprintf(outf,"  ");
  */
  i=strlen(outs);
  if (i>5)
    if (!strcmp(outs+i-5,"[R00]"))
      outs[i-5]=0;
  sprintf(tmp,"%s%s",outm,outs); 
	}
else
  sprintf(tmp,"%02X%27s",*p,"unimplemented"); 
strcpy(outs,tmp);
return n+prefix;
}

int instr(Word PC)
{
int len,i;
len=disassemble(PC);
printsym(PC);
for (i=0;i<5;i++)
  if (i<len)
		printf("%02X",*(mem+PC+i));
	else
	   printf("  ");
		printf(" %s\n",outs);
		return len;
}
						
